// nearest.cpp
// MCS 481 project 3 description version 1.0
// David Dumas
//
// This is a modified version of a CGAL demo:
//   examples/Voronoi_diagram_2/vd_2_point_location.cpp

#include <iostream>
#include <fstream>

// includes for defining the Voronoi diagram adaptor
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Delaunay_triangulation_2.h>
#include <CGAL/Voronoi_diagram_2.h>
#include <CGAL/Delaunay_triangulation_adaptation_traits_2.h>
#include <CGAL/Delaunay_triangulation_adaptation_policies_2.h>

// typedefs for defining the adaptor
typedef CGAL::Exact_predicates_inexact_constructions_kernel                  K;
typedef CGAL::Delaunay_triangulation_2<K>                                    DT;
typedef CGAL::Delaunay_triangulation_adaptation_traits_2<DT>                 AT;
typedef CGAL::Delaunay_triangulation_caching_degeneracy_removal_policy_2<DT> AP;
typedef CGAL::Voronoi_diagram_2<DT,AT,AP>                                    VD;

// typedef for the result type of the point location
typedef AT::Site_2                    Site_2;
typedef AT::Point_2                   Point_2;

typedef VD::Locate_result             Locate_result;
typedef VD::Vertex_handle             Vertex_handle;
typedef VD::Delaunay_vertex_handle    Delaunay_vertex_handle;
typedef VD::Face_handle               Face_handle;
typedef VD::Halfedge_handle           Halfedge_handle;
typedef VD::Ccb_halfedge_circulator   Ccb_halfedge_circulator;

#include <ctime>
#include <string>
// global time variable
clock_t LAST;

void tic()
{
  LAST = clock();
}

void toc(std::string s)
{
  double t = (double)(clock() - LAST) / (double)CLOCKS_PER_SEC;
  std::cerr << std::fixed << t << " seconds " << s << std::endl;
}

void print_endpoint(Halfedge_handle e, bool is_src)
{
  std::cout << "\t";
  if ( is_src ) {
    if ( e->has_source() )  std::cout << e->source()->point() << std::endl;
    else  std::cout << "point at infinity" << std::endl;
  } else {
    if ( e->has_target() )  std::cout << e->target()->point() << std::endl;
    else  std::cout << "point at infinity" << std::endl;
  }
}

int main(int argc, char **argv)
{
  bool quiet = false;
  int idx = 1;

  if ((argc > 1) && (!strcmp(argv[1],"-q"))) {
    quiet = true;
    idx++;
  }

  if (argc < (idx+1)) {
    std::cerr << "ERROR: No site file specified." << std::endl;
    exit(1);
  }

  // sitefile
  std::ifstream sitefile(argv[idx]);
  
  // Construct the voronoi diagram

  tic();
  VD vd;
  Site_2 t;
  while ( sitefile >> t ) { vd.insert(t); }
  sitefile.close();
  toc("voronoi construction");

  tic();
  // Perform point location queries
  Point_2 p;
  while ( std::cin >> p ) {
    Locate_result lr = vd.locate(p);
    if (quiet) {
      // Simple output
      Delaunay_vertex_handle s;
      if ( Vertex_handle* v = boost::get<Vertex_handle>(&lr) ) {
	s = (*v)->site(0);
      } else if ( Halfedge_handle* e = boost::get<Halfedge_handle>(&lr) ) {
	s = (*e)->up();
      } else if (Face_handle* f = boost::get<Face_handle>(&lr) ) {
	s = (*f)->dual();
      }      
      std::cout << s->point() << std::endl;
    } else {
      // Full descriptive output
      std::cout << "Query point (" << p.x() << "," << p.y()
		<< ") lies on a Voronoi " << std::flush;
      
      if ( Vertex_handle* v = boost::get<Vertex_handle>(&lr) ) {
	std::cout << "vertex." << std::endl;
	std::cout << "The Voronoi vertex is:" << std::endl;
	std::cout << "\t" << (*v)->point() << std::endl;
      } else if ( Halfedge_handle* e = boost::get<Halfedge_handle>(&lr) ) {
	std::cout << "edge." << std::endl;
	std::cout << "The source and target vertices "
		  << "of the Voronoi edge are:" << std::endl;
	print_endpoint(*e, true);
	print_endpoint(*e, false);
      } else if ( Face_handle* f = boost::get<Face_handle>(&lr) ) {
	std::cout << "face." << std::endl;
	std::cout << "The vertices of the Voronoi face are"
		  << " (in counterclockwise order):" << std::endl;
	Ccb_halfedge_circulator ec_start = (*f)->outer_ccb();
	Ccb_halfedge_circulator ec = ec_start;
	do {
	  print_endpoint(ec, false);
	} while ( ++ec != ec_start );
      }
      std::cout << std::endl;
    }
  }

  toc("point location");
  return 0;
}
